import logging
import re
import time
from logging import Logger
from typing import Any, Callable, Dict, List, Type
from urllib.parse import SplitResult, urlsplit

from bs4 import BeautifulSoup
from pydantic import AnyHttpUrl, parse_obj_as
from requests import PreparedRequest, Request, Response
from requests.adapters import HTTPAdapter

from .__about__ import __version__
from .exceptions import (
    ParseModelError,
    ServerHtmlError,
    ServerJsonError,
    ServerMultiJsonError,
    ServerTextError,
)
from .model.commons import (
    Error,
    MsData,
    MsDataUrl,
    MsError,
    MsStatus,
    SaeAlert,
    TiAlert,
)
from .model.enums import Api, HttpMethod, Provider, Status
from .model.requests import EndpointTask
from .model.responses import (
    MR,
    AddAlertNoteResp,
    BaseLinkableResp,
    BaseMultiResponse,
    BytesResp,
    C,
    ConsumeLinkableResp,
    GetAlertDetailsResp,
    MultiResp,
    MultiUrlResp,
    NoContentResp,
    R,
    S,
    SandboxSubmissionStatusResp,
)
from .results import multi_result, result

USERAGENT_SUFFIX: str = "PyTMV1"
API_VERSION: str = "v3.0"

log: Logger = logging.getLogger(__name__)


class Core:
    def __init__(
        self,
        appname: str,
        token: str,
        url: str,
        pool_connections: int,
        pool_maxsize: int,
        connect_timeout: int,
        read_timeout: int,
    ):
        self._adapter = HTTPAdapter(pool_connections, pool_maxsize, 0, True)
        self._c_timeout = connect_timeout
        self._r_timeout = read_timeout
        self._appname = appname
        self._token = token
        self._url = parse_obj_as(AnyHttpUrl, _format(url))
        self._headers: Dict[str, str] = {
            "Authorization": f"Bearer {self._token}",
            "User-Agent": f"{self._appname}-{USERAGENT_SUFFIX}/{__version__}",
        }

    @result
    def send(
        self,
        class_: Type[R],
        api: str,
        method: HttpMethod = HttpMethod.GET,
        **kwargs: Any,
    ) -> R:
        return self._process(
            class_,
            api,
            method,
            **kwargs,
        )

    @multi_result
    def send_endpoint(
        self,
        api: Api,
        *tasks: EndpointTask,
    ) -> MultiResp:
        return self._process(
            MultiResp,
            api,
            HttpMethod.POST,
            json=[
                task.dict(by_alias=True, exclude_none=True) for task in tasks
            ],
        )

    @result
    def send_linkable(
        self,
        class_: Type[BaseLinkableResp[C]],
        api: str,
        consumer: Callable[[C], None],
        **kwargs: Any,
    ) -> ConsumeLinkableResp:
        return ConsumeLinkableResp(
            total_consumed=self._consume_linkable(
                lambda: self._process(
                    class_,
                    api,
                    **kwargs,
                ),
                consumer,
            )
        )

    @multi_result
    def send_multi(
        self,
        class_: Type[MR],
        api: str,
        **kwargs: Any,
    ) -> MR:
        return self._process(
            class_,
            api,
            HttpMethod.POST,
            **kwargs,
        )

    @result
    def send_sandbox_result(
        self,
        class_: Type[R],
        api: Api,
        submit_id: str,
        poll: bool,
        poll_time_sec: float,
    ) -> R:
        if poll:
            _poll_status(
                lambda: self._process(
                    SandboxSubmissionStatusResp,
                    Api.GET_SANDBOX_SUBMISSION_STATUS.value.format(submit_id),
                ),
                poll_time_sec,
            )
        return self._process(class_, api.value.format(submit_id))

    @result
    def send_task_result(
        self, class_: Type[S], task_id: str, poll: bool, poll_time_sec: float
    ) -> S:
        status_call: Callable[[], S] = lambda: self._process(
            class_,
            Api.GET_TASK_RESULT.value.format(task_id),
        )
        if poll:
            _poll_status(
                status_call,
                poll_time_sec,
            )
        return status_call()

    def _consume_linkable(
        self,
        api_call: Callable[[], BaseLinkableResp[C]],
        consumer: Callable[[C], None],
        count: int = 0,
    ) -> int:
        total_count: int = count
        response: BaseLinkableResp[C] = api_call()
        for item in response.items:
            consumer(item)
            total_count += 1
        if response.next_link:
            sr: SplitResult = urlsplit(response.next_link)
            log.info("Found nextLink")
            return self._consume_linkable(
                lambda: self._process(
                    type(response),
                    sr.path[5:],
                    params={"skipToken": sr.query.split("=")[-1]},
                ),
                consumer,
                total_count,
            )
        log.info(
            "Records consumed: [Total=%s, Type=%s]",
            total_count,
            type(
                response.items[0] if len(response.items) > 0 else response
            ).__name__,
        )
        return total_count

    def _process(
        self,
        class_: Type[R],
        uri: str,
        method: HttpMethod = HttpMethod.GET,
        **kwargs: Any,
    ) -> R:
        log.info(
            "Processing request [Method=%s, Class=%s, Api=%s, Options=%s]",
            method.value,
            class_.__name__,
            uri,
            kwargs,
        )
        raw_response: Response = self._send_internal(
            self._prepare(uri, method, **kwargs)
        )
        _validate(raw_response)
        return _parse_data(raw_response, class_)

    def _prepare(
        self, uri: str, method: HttpMethod, **kwargs: Any
    ) -> PreparedRequest:
        return Request(
            method.value,
            self._url + uri,
            headers={**self._headers, **kwargs.pop("headers", {})},
            **kwargs,
        ).prepare()

    def _send_internal(self, request: PreparedRequest) -> Response:
        log.info(
            "Sending request [Method=%s, URL=%s, Headers=%s, Body=%s]",
            request.method,
            request.url,
            re.sub("Bearer [^\\s']+", "*****", str(request.headers)),
            (
                request.body.decode("utf-8")
                if type(request.body) == bytes
                else request.body
            ),
        )
        response: Response = self._adapter.send(
            request, timeout=(self._c_timeout, self._r_timeout)
        )
        log.info(
            "Received response [Status=%s, Headers=%s, Body=%s]",
            response.status_code,
            response.headers,
            _hide_binary(response),
        )
        return response


def _format(url: str) -> str:
    return (url if url.endswith("/") else url + "/") + API_VERSION


def _hide_binary(response: Response) -> str:
    content_type = response.headers.get("Content-Type", "")
    if "json" not in content_type and "application" in content_type:
        return "***binary content***"
    return response.text


def _is_http_success(status_codes: List[int]) -> bool:
    return len(list(filter(lambda s: not 200 <= s < 399, status_codes))) == 0


def _parse_data(raw_response: Response, class_: Type[R]) -> R:
    content_type = raw_response.headers.get("Content-Type", "")
    if "json" in content_type:
        if issubclass(class_, BaseMultiResponse):
            log.info("Parsing json multi response [Class=%s]", class_.__name__)
            class_d: Type[List[Any]]
            if issubclass(class_, MultiUrlResp):
                class_d = List[MsDataUrl]
            else:
                class_d = List[MsData]
            return class_(
                items=parse_obj_as(
                    class_d,
                    raw_response.json(),
                )
            )
        log.info("Parsing json response [Class=%s]", class_.__name__)
        if class_ == GetAlertDetailsResp:
            response_json: Dict[str, str] = raw_response.json()
            return class_(
                alert=parse_obj_as(
                    (
                        SaeAlert
                        if response_json.get("alertProvider") == Provider.SAE
                        else TiAlert
                    ),
                    response_json,
                ),
                etag=raw_response.headers.get("ETag", ""),
            )
        return class_.parse_obj(raw_response.json())
    if "application" in content_type and class_ == BytesResp:
        log.info("Parsing binary response")
        return class_(content=raw_response.content)
    if raw_response.status_code == 201 and class_ == AddAlertNoteResp:
        return class_.parse_obj(raw_response.headers)
    if raw_response.status_code == 204 and class_ == NoContentResp:
        return class_()
    raise ParseModelError(class_.__name__, raw_response)


def _parse_html(html: str) -> str:
    log.info("Parsing html response [Html=%s]", html)
    soup = BeautifulSoup(html, "html.parser")
    return "\n".join(
        line.strip() for line in soup.text.split("\n") if line.strip()
    )


def _poll_status(
    status_call: Callable[[], S],
    poll_time_sec: float,
) -> None:
    start_time: float = time.time()
    elapsed_time: float = 0
    response: S = status_call()
    while elapsed_time < poll_time_sec:
        if response.status in [Status.QUEUED, Status.RUNNING]:
            response = status_call()
            elapsed_time = time.time() - start_time
        else:
            break


def _validate(raw_response: Response) -> None:
    log.info("Validating response [%s]", raw_response)
    content_type: str = raw_response.headers.get("Content-Type", "")
    if "text/html" in content_type:
        raise ServerHtmlError(
            raw_response.status_code, _parse_html(raw_response.text)
        )
    if not _is_http_success([raw_response.status_code]):
        if "application/json" in content_type:
            error: Dict[str, Any] = raw_response.json().get("error")
            error["status"] = raw_response.status_code
            raise ServerJsonError(
                Error.parse_obj(error),
            )
        raise ServerTextError(raw_response.status_code, raw_response.text)
    if raw_response.status_code == 207:
        if not _is_http_success(
            MsStatus.parse_obj(raw_response.json()).values()
        ):
            raise ServerMultiJsonError(
                parse_obj_as(List[MsError], raw_response.json())
            )



from typing import Any, Dict, Generic, List, Optional, Tuple, TypeVar, Union

from pydantic import Field
from pydantic.generics import GenericModel

from .commons import (
    Account,
    BaseConsumable,
    BaseModel,
    Digest,
    EmailMessage,
    Endpoint,
    ExceptionObject,
    MsData,
    MsDataUrl,
    SaeAlert,
    SandboxSuspiciousObject,
    SuspiciousObject,
    TiAlert,
)
from .enums import (
    ObjectType,
    RiskLevel,
    SandboxAction,
    SandboxObjectType,
    Status,
    TaskAction,
)

C = TypeVar("C", bound=BaseConsumable)
M = TypeVar("M", bound=MsData)


class BaseResponse(BaseModel):
    ...


class BaseLinkableResp(BaseResponse, GenericModel, Generic[C]):
    next_link: Optional[str]
    items: List[C] = []


class BaseMultiResponse(BaseResponse, GenericModel, Generic[M]):
    items: List[M] = []


class BaseStatusResponse(BaseResponse):
    id: str
    status: Status
    created_date_time: str
    last_action_date_time: str


class BaseTaskResp(BaseStatusResponse):
    action: TaskAction
    description: Optional[str]
    account: Optional[str]


MR = TypeVar("MR", bound=BaseMultiResponse[Any])
R = TypeVar("R", bound=BaseResponse)
S = TypeVar("S", bound=BaseStatusResponse)


class AccountTaskResp(BaseTaskResp):
    tasks: List[Account]


class AddAlertNoteResp(BaseResponse):
    location: str = Field(alias="Location")

    def note_id(self) -> str:
        return self.location.split("/")[-1]


class BlockListTaskResp(BaseTaskResp):
    type: ObjectType
    value: str

    def __init__(self, **data: str) -> None:
        obj: Tuple[str, str] = self._map(data)
        super().__init__(type=obj[0], value=obj[1], **data)

    @staticmethod
    def _map(args: Dict[str, str]) -> Tuple[str, str]:
        return {
            (k, v)
            for k, v in args.items()
            if k in map(lambda ot: ot.value, ObjectType)
        }.pop()


class BytesResp(BaseResponse):
    content: bytes


class CollectFileTaskResp(BaseTaskResp):
    agent_guid: str
    endpoint_name: str
    file_path: str
    file_sha1: Optional[str]
    file_sha256: Optional[str]
    file_size: Optional[int]
    resource_location: Optional[str]
    expired_date_time: Optional[str]
    password: Optional[str]


class ConnectivityResp(BaseResponse):
    status: str


class ConsumeLinkableResp(BaseResponse, alias_generator=None):
    total_consumed: int


class EndpointTaskResp(BaseTaskResp):
    agent_guid: str
    endpoint_name: str


class GetAlertDetailsResp(BaseResponse):
    alert: Union[SaeAlert, TiAlert]
    etag: str


class GetAlertListResp(BaseLinkableResp[Union[SaeAlert, TiAlert]]):
    total_count: int
    count: int


class GetEndpointDataResp(BaseLinkableResp[Endpoint]):
    ...


class GetExceptionListResp(BaseLinkableResp[ExceptionObject]):
    ...


class GetSuspiciousListResp(BaseLinkableResp[SuspiciousObject]):
    ...


class MultiResp(BaseMultiResponse[MsData]):
    ...


class MultiUrlResp(BaseMultiResponse[MsDataUrl]):
    ...


class NoContentResp(BaseResponse):
    ...


class EmailMessageTaskResp(BaseTaskResp):
    tasks: List[EmailMessage]


class SubmitFileToSandboxResp(BaseResponse):
    id: str
    digest: Digest
    arguments: Optional[str]


class SandboxAnalysisResultResp(BaseResponse):
    id: str
    type: SandboxObjectType
    analysis_completion_date_time: str
    risk_level: RiskLevel
    true_file_type: Optional[str]
    digest: Optional[Digest]
    arguments: Optional[str]
    detection_names: List[str] = Field(default=List)
    threat_types: List[str] = Field(default=List)


class SandboxSubmissionStatusResp(BaseStatusResponse):
    action: SandboxAction
    resource_location: Optional[str]
    is_cached: Optional[bool]
    digest: Optional[Digest]
    arguments: Optional[str]


class SandboxSuspiciousListResp(BaseResponse):
    items: List[SandboxSuspiciousObject]


class SandboxSubmitUrlTaskResp(BaseTaskResp):
    url: str
    sandbox_task_id: str


class TerminateProcessTaskResp(BaseTaskResp):
    agent_guid: str
    endpoint_name: str
    file_sha1: str
    file_name: Optional[str]



from enum import Enum


class Api(str, Enum):
    ADD_ALERT_NOTE = "/workbench/alerts/{0}/notes"
    ADD_TO_BLOCK_LIST = "/response/suspiciousObjects"
    ADD_TO_EXCEPTION_LIST = "/threatintel/suspiciousObjectExceptions"
    ADD_TO_SUSPICIOUS_LIST = "/threatintel/suspiciousObjects"
    COLLECT_ENDPOINT_FILE = "/response/endpoints/collectFile"
    CONNECTIVITY = ("/healthcheck/connectivity",)
    DELETE_EMAIL_MESSAGE = "/response/emails/delete"
    DISABLE_ACCOUNT = "/response/domainAccounts/disable"
    DOWNLOAD_SANDBOX_ANALYSIS_RESULT = "/sandbox/analysisResults/{0}/report"
    DOWNLOAD_SANDBOX_INVESTIGATION_PACKAGE = (
        "/sandbox/analysisResults/{0}/investigationPackage"
    )
    EDIT_ALERT_STATUS = "/workbench/alerts/{0}"
    ENABLE_ACCOUNT = "/response/domainAccounts/enable"
    ISOLATE_ENDPOINT = "/response/endpoints/isolate"
    GET_ALERT_DETAILS = "/workbench/alerts/{0}"
    GET_ALERT_LIST = "/workbench/alerts"
    GET_ENDPOINT_DATA = "/eiqs/endpoints"
    GET_EXCEPTION_LIST = "/threatintel/suspiciousObjectExceptions"
    GET_SANDBOX_SUBMISSION_STATUS = "/sandbox/tasks/{0}"
    GET_SANDBOX_ANALYSIS_RESULT = "/sandbox/analysisResults/{0}"
    GET_SANDBOX_SUSPICIOUS_LIST = (
        "/sandbox/analysisResults/{0}/suspiciousObjects"
    )
    GET_SUSPICIOUS_LIST = "/threatintel/suspiciousObjects"
    GET_TASK_RESULT = "/response/tasks/{0}"
    QUARANTINE_EMAIL_MESSAGE = "/response/emails/quarantine"
    REMOVE_FROM_BLOCK_LIST = "/response/suspiciousObjects/delete"
    REMOVE_FROM_EXCEPTION_LIST = (
        "/threatintel/suspiciousObjectExceptions/delete"
    )
    REMOVE_FROM_SUSPICIOUS_LIST = "/threatintel/suspiciousObjects/delete"
    RESET_PASSWORD = "/response/domainAccounts/resetPassword"
    RESTORE_EMAIL_MESSAGE = "/response/emails/restore"
    RESTORE_ENDPOINT = "/response/endpoints/restore"
    SIGN_OUT_ACCOUNT = "/response/domainAccounts/signOut"
    SUBMIT_FILE_TO_SANDBOX = "/sandbox/files/analyze"
    SUBMIT_URLS_TO_SANDBOX = "/sandbox/urls/analyze"
    TERMINATE_ENDPOINT_PROCESS = "/response/endpoints/terminateProcess"


class Iam(str, Enum):
    AAD = "Azure AD"
    OPAD = "On-premise AD"


class InvestigationStatus(str, Enum):
    BENIGN_TRUE_POSITIVE = "Benign True Positive"
    CLOSED = "Closed"
    FALSE_POSITIVE = "False Positive"
    IN_PROGRESS = "In Progress"
    NEW = "New"
    TRUE_POSITIVE = "True Positive"


class EntityType(str, Enum):
    HOST = "host"
    ACCOUNT = "account"
    EMAIL_ADDRESS = "emailAddress"


class HttpMethod(str, Enum):
    GET = "GET"
    PATCH = "PATCH"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"


class ObjectType(str, Enum):
    IP = "ip"
    URL = "url"
    DOMAIN = "domain"
    FILE_SHA1 = "fileSha1"
    FILE_SHA256 = "fileSha256"
    SENDER_MAIL_ADDRESS = "senderMailAddress"


class OperatingSystem(str, Enum):
    LINUX = "Linux"
    WINDOWS = "Windows"
    MACOS = "macOS"
    MACOSX = "macOSX"


class ProductCode(str, Enum):
    SAO = "sao"
    SDS = "sds"
    XES = "xes"


class Provenance(str, Enum):
    ALERT = "Alert"
    SWEEPING = "Sweeping"
    NETWORK_ANALYTICS = "Network Analytics"


class Provider(str, Enum):
    SAE = "SAE"
    TI = "TI"


class QueryField(str, Enum):
    AGENT_GUID = "agentGuid"
    LOGIN_ACCOUNT = "loginAccount"
    ENDPOINT_NAME = "endpointName"
    MAC_ADDRESS = "macAddress"
    IP = "ip"
    OS_NAME = "osName"
    PRODUCT_CODE = "productCode"
    INSTALLED_PRODUCT_CODES = "installedProductCodes"


class QueryOp(str, Enum):
    AND = " and "
    OR = " or "


class RiskLevel(str, Enum):
    NO_RISK = "noRisk"
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"


class SandboxAction(str, Enum):
    ANALYZE_FILE = "analyzeFile"
    ANALYZE_URL = "analyzeUrl"


class SandboxObjectType(str, Enum):
    URL = "url"
    FILE = "file"


class ScanAction(str, Enum):
    BLOCK = "block"
    LOG = "log"


class Severity(str, Enum):
    CRITICAL = "critical"
    HIGH = "high"
    MEDIUM = "medium"
    LOW = "low"


class Status(str, Enum):
    FAILED = "failed"
    QUEUED = "queued"
    REJECTED = "rejected"
    RUNNING = "running"
    SUCCEEDED = "succeeded"
    WAIT_FOR_APPROVAL = "waitForApproval"


class TaskAction(str, Enum):
    COLLECT_FILE = ("collectFile",)
    ISOLATE_ENDPOINT = ("isolate",)
    RESTORE_ENDPOINT = ("restoreIsolate",)
    TERMINATE_PROCESS = ("terminateProcess",)
    QUARANTINE_MESSAGE = "quarantineMessage"
    DELETE_MESSAGE = ("deleteMessage",)
    RESTORE_MESSAGE = ("restoreMessage",)
    BLOCK_SUSPICIOUS = ("block",)
    REMOVE_SUSPICIOUS = ("restoreBlock",)
    RESET_PASSWORD = "resetPassword"
    SUBMIT_SANDBOX = ("submitSandbox",)
    ENABLE_ACCOUNT = ("enableAccount",)
    DISABLE_ACCOUNT = ("disableAccount",)
    FORCE_SIGN_OUT = "forceSignOut"



import base64
import re
from typing import Any, Dict, List, Optional, Pattern, Tuple

from pydantic import IPvAnyAddress, IPvAnyAddressError

from .model.enums import OperatingSystem, ProductCode, QueryField, QueryOp
from .model.requests import ObjectTask, SuspiciousObjectTask

MAC_ADDRESS_PATTERN: Pattern[str] = re.compile(
    "^([0-9A-Fa-f]{2}[:-]){5}([0-9A-Fa-f]{2})$"
)
GUID_PATTERN: Pattern[str] = re.compile("^(\\w+-+){1,5}\\w+$")


def build_object_request(*tasks: ObjectTask) -> List[Dict[str, str]]:
    return [
        filter_none(
            {
                task.object_type.value: task.object_value,
                "description": task.description,
            }
        )
        for task in tasks
    ]


def build_sandbox_file_request(
    document_password: Optional[str],
    archive_password: Optional[str],
    arguments: Optional[str],
) -> Dict[str, str]:
    return filter_none(
        {
            "documentPassword": _b64_encode(document_password),
            "archivePassword": _b64_encode(archive_password),
            "arguments": _b64_encode(arguments),
        }
    )


def build_suspicious_request(
    *tasks: SuspiciousObjectTask,
) -> List[Dict[str, Any]]:
    return [
        filter_none(
            {
                task.object_type.value: task.object_value,
                "description": task.description,
                "riskLevel": (
                    task.risk_level.value if task.risk_level else None
                ),
                "scanAction": (
                    task.scan_action.value if task.scan_action else None
                ),
                "daysToExpiration": task.days_to_expiration,
            }
        )
        for task in tasks
    ]


def endpoint_query(op: QueryOp, *values: str) -> Dict[str, str]:
    return {
        "TMV1-Query": op.join(
            "("
            + QueryOp.OR.join(
                f"{qt.value} eq '{value}'"
                for qt in endpoint_query_field(value)
            )
            + ")"
            for value in values
        )
    }


def endpoint_query_field(value: str) -> Tuple[QueryField, ...]:
    if _is_ip_address(value):
        return (QueryField.IP,)
    if bool(MAC_ADDRESS_PATTERN.match(value)):
        return (QueryField.MAC_ADDRESS,)
    if bool(GUID_PATTERN.match(value)):
        return (QueryField.AGENT_GUID,)
    if next(filter(lambda os: os.value == value, OperatingSystem), None):
        return (QueryField.OS_NAME,)
    if next(filter(lambda pc: pc.value == value, ProductCode), None):
        return QueryField.PRODUCT_CODE, QueryField.INSTALLED_PRODUCT_CODES
    return QueryField.ENDPOINT_NAME, QueryField.LOGIN_ACCOUNT


def filter_none(dictionary: Dict[str, Optional[Any]]) -> Dict[str, Any]:
    return {k: v for k, v in dictionary.items() if v}


def _b64_encode(value: Optional[str]) -> Optional[str]:
    return base64.b64encode(value.encode()).decode() if value else None


def _is_ip_address(endpoint_value: str) -> bool:
    try:
        return bool(IPvAnyAddress.validate(endpoint_value))
    except IPvAnyAddressError:
        return False



from __future__ import annotations

from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel as PydanticBaseModel
from pydantic.utils import to_lower_camel

from .enums import (
    EntityType,
    Iam,
    InvestigationStatus,
    ObjectType,
    OperatingSystem,
    ProductCode,
    Provider,
    RiskLevel,
    ScanAction,
    Severity,
    Status,
)


class BaseModel(PydanticBaseModel):
    class Config:
        alias_generator = to_lower_camel


class BaseConsumable(BaseModel):
    ...


def _get_task_id(headers: List[Dict[str, str]]) -> Optional[str]:
    task_id: str = next(
        (
            h.get("value", "")
            for h in headers
            if "Operation-Location" == h.get("name", "")
        ),
        "",
    ).split("/")[-1]
    return task_id if task_id != "" else None


class Account(BaseModel):
    account_name: str
    iam: Iam
    last_action_date_time: str
    status: Status


class Alert(BaseConsumable):
    id: str
    schema_version: str
    investigation_status: InvestigationStatus
    workbench_link: str
    alert_provider: Provider
    model: str
    score: int
    severity: Severity
    impact_scope: ImpactScope
    created_date_time: str
    updated_date_time: str
    indicators: List[Indicator]


class Digest(BaseModel):
    md5: str
    sha1: str
    sha256: str


class EmailMessage(BaseModel):
    last_action_date_time: str
    message_id: Optional[str]
    mail_box: Optional[str]
    message_subject: Optional[str]
    unique_id: Optional[str]
    organization_id: Optional[str]
    status: Status


class Value(BaseModel):
    updated_date_time: str
    value: str


class ValueList(BaseModel):
    updated_date_time: str
    value: List[str]


class Endpoint(BaseConsumable):
    agent_guid: str
    login_account: ValueList
    endpoint_name: Value
    mac_address: ValueList
    ip: ValueList
    os_name: OperatingSystem
    os_version: str
    os_description: str
    product_code: ProductCode
    installed_product_codes: List[ProductCode]


class HostInfo(BaseModel):
    name: str
    ips: List[str]
    guid: str


class Entity(BaseModel):
    entity_id: str
    entity_type: EntityType
    entity_value: Union[str, HostInfo]
    related_entities: List[str]
    related_indicator_ids: List[int]
    provenance: List[str]


class Error(BaseModel):
    status: int
    code: Optional[str] = None
    message: Optional[str] = None
    number: Optional[int] = None


class ExceptionObject(BaseConsumable):
    value: str
    type: ObjectType
    last_modified_date_time: str
    description: Optional[str]

    def __init__(self, **data: str) -> None:
        super().__init__(value=self._obj_value(data), **data)

    @staticmethod
    def _obj_value(args: Dict[str, str]) -> str:
        obj_value: Optional[str] = args.get(args.get("type", ""))
        if obj_value is None:
            raise ValueError("Object value not found")
        return obj_value


class ImpactScope(BaseModel):
    desktop_count: int
    server_count: int
    account_count: int
    email_address_count: int
    entities: List[Entity]


class Indicator(BaseModel):
    id: int
    type: str
    value: Union[str, HostInfo]
    related_entities: List[str]
    provenance: List[str]


class MatchedEvent(BaseModel):
    uuid: str
    matched_date_time: str
    type: str


class MatchedFilter(BaseModel):
    id: str
    name: str
    matched_date_time: str
    mitre_technique_ids: List[str]
    matched_events: List[MatchedEvent]


class MatchedIndicatorPattern(BaseModel):
    id: str
    pattern: str
    tags: List[str]
    matched_logs: List[str]


class MatchedRule(BaseModel):
    id: str
    name: str
    matched_filters: List[MatchedFilter]


class MsData(BaseModel):
    status: int
    task_id: Optional[str] = None

    def __init__(self, **data: Any):
        super().__init__(
            taskId=_get_task_id(data.pop("headers", {})),
            **data,
        )


class MsDataUrl(MsData):
    url: str
    id: Optional[str]
    digest: Optional[Digest]

    def __init__(self, **data: Any):
        data.update(data.pop("body", {}))
        super().__init__(**data)


class MsError(Error):
    extra: Dict[str, str] = {}
    task_id: Optional[str]

    def __init__(self, **data: Any):
        data.update(data.pop("body", {}))
        data.update(data.pop("error", {}))
        super().__init__(
            extra={"url": data.pop("url", "")},
            taskId=_get_task_id(data.pop("headers", {})),
            **data,
        )


class MsStatus(BaseModel):
    __root__: List[int]

    def __init__(self, **data: Any):
        super().__init__(
            root=[int(d.get("status", 500)) for d in data.get("__root__", [])]
        )

    def values(self) -> List[int]:
        return self.__root__


class SaeAlert(Alert):
    description: str
    matched_rules: List[MatchedRule]


class SaeIndicator(Indicator):
    field: str
    filter_ids: List[str]


class SandboxSuspiciousObject(BaseModel):
    risk_level: RiskLevel
    analysis_completion_date_time: str
    expired_date_time: str
    root_sha1: str
    type: ObjectType
    value: str

    def __init__(self, **data: str) -> None:
        obj: Tuple[str, str] = self._map(data)
        super().__init__(type=obj[0], value=obj[1], **data)

    @staticmethod
    def _map(args: Dict[str, str]) -> Tuple[str, str]:
        return {
            (k, v)
            for k, v in args.items()
            if k in map(lambda ot: ot.value, ObjectType)
        }.pop()


class SuspiciousObject(ExceptionObject):
    scan_action: ScanAction
    risk_level: RiskLevel
    in_exception_list: bool
    expired_date_time: str


class TiAlert(Alert):
    campaign: Optional[str]
    industry: Optional[str]
    region_and_country: Optional[str]
    created_by: str
    total_indicator_count: int
    matched_indicator_count: int
    report_link: str
    matched_indicator_patterns: List[MatchedIndicatorPattern]


class TiIndicator(Indicator):
    fields: List[List[str]]
    matched_indicator_pattern_ids: List[str]
    first_seen_date_times: List[str]
    last_seen_date_times: List[str]



from __future__ import annotations

import logging
import time
from dataclasses import dataclass, field
from enum import Enum
from functools import wraps
from logging import Logger
from typing import Any, Callable, Generic, List, Optional, TypeVar

from pydantic import ValidationError
from requests import RequestException

from .exceptions import ServerCustError, ServerJsonError, ServerMultiJsonError
from .model.commons import Error, MsError
from .model.responses import MR, R

E = TypeVar("E", bound=Error)
F = TypeVar("F", bound=Callable[..., Any])

log: Logger = logging.getLogger(__name__)


def multi_result(func: F) -> Callable[..., MultiResult[MR]]:
    @wraps(func)
    def _multi_result(*args: Any, **kwargs: Any) -> MultiResult[MR]:
        obj: MR | Exception = _wrapper(func, *args, **kwargs)
        return (
            MultiResult.success(obj)
            if not isinstance(obj, Exception)
            else MultiResult.failed(obj)
        )

    return _multi_result


def result(func: F) -> Callable[..., Result[R]]:
    @wraps(func)
    def _result(*args: Any, **kwargs: Any) -> Result[R]:
        obj: R | Exception = _wrapper(func, *args, **kwargs)
        return (
            Result.success(obj)
            if not isinstance(obj, Exception)
            else Result.failed(obj)
        )

    return _result


def _wrapper(func: F, *args: Any, **kwargs: Any) -> R | Exception:
    try:
        start_time: float = time.time()
        log.debug(
            "Execution started [%s, %s]",
            args,
            kwargs,
        )
        response: R = func(*args, **kwargs)
        log.debug(
            "Execution finished [Elapsed=%s, %s]",
            time.time() - start_time,
            response,
        )
        return response
    except (
        ServerCustError,
        ServerJsonError,
        ServerMultiJsonError,
        ValidationError,
        RequestException,
        RuntimeError,
    ) as exc:
        log.exception("Unexpected issue occurred [%s]", exc)
        return exc


def _error(exc: Exception) -> Error:
    if isinstance(exc, ServerJsonError):
        return exc.error
    return Error(
        status=_status(exc), code=type(exc).__name__, message=str(exc)
    )


def _errors(exc: Exception) -> List[MsError]:
    if isinstance(exc, ServerMultiJsonError):
        return exc.errors
    if isinstance(exc, ServerJsonError):
        return [
            MsError(
                status=exc.error.status,
                code=exc.error.code,
                message=exc.error.message,
                number=exc.error.number,
            )
        ]
    return [
        MsError(status=_status(exc), code=type(exc).__name__, message=str(exc))
    ]


def _status(exc: Exception) -> int:
    return exc.status if isinstance(exc, ServerCustError) else 500


@dataclass
class BaseResult(Generic[R]):
    result_code: ResultCode
    response: Optional[R] = None


@dataclass
class Result(BaseResult[R]):
    error: Optional[Error] = None

    @classmethod
    def success(cls, response: R) -> Result[R]:
        return cls(ResultCode.SUCCESS, response)

    @classmethod
    def failed(cls, exc: Exception) -> Result[R]:
        return cls(
            ResultCode.ERROR,
            None,
            _error(exc),
        )


@dataclass
class MultiResult(BaseResult[MR]):
    errors: List[MsError] = field(default_factory=list)

    @classmethod
    def success(cls, response: MR) -> MultiResult[MR]:
        return cls(ResultCode.SUCCESS, response)

    @classmethod
    def failed(cls, exc: Exception) -> MultiResult[MR]:
        return cls(
            ResultCode.ERROR,
            None,
            _errors(exc),
        )


class ResultCode(str, Enum):
    SUCCESS = "SUCCESS"
    ERROR = "ERROR"



from __future__ import annotations

import logging
from functools import lru_cache
from logging import Logger
from typing import Callable, Optional, Type, Union

from . import utils
from .core import Core
from .model.commons import (
    Endpoint,
    ExceptionObject,
    SaeAlert,
    SuspiciousObject,
    TiAlert,
)
from .model.enums import Api, HttpMethod, InvestigationStatus, QueryOp
from .model.requests import (
    AccountTask,
    EmailMessageIdTask,
    EmailMessageUIdTask,
    EndpointTask,
    FileTask,
    ObjectTask,
    ProcessTask,
    SuspiciousObjectTask,
)
from .model.responses import (
    AddAlertNoteResp,
    BaseTaskResp,
    BytesResp,
    ConnectivityResp,
    ConsumeLinkableResp,
    GetAlertDetailsResp,
    GetAlertListResp,
    GetEndpointDataResp,
    GetExceptionListResp,
    GetSuspiciousListResp,
    MultiResp,
    MultiUrlResp,
    NoContentResp,
    S,
    SandboxAnalysisResultResp,
    SandboxSubmissionStatusResp,
    SandboxSuspiciousListResp,
    SubmitFileToSandboxResp,
)
from .results import MultiResult, Result

log: Logger = logging.getLogger(__name__)


def client(
    name: str,
    token: str,
    url: str,
    pool_connections: int = 1,
    pool_maxsize: int = 1,
    connect_timeout: int = 30,
    read_timeout: int = 30,
) -> Client:
    """Helper function to initialize a :class:`Client`.

    :param name: Identify the application using this library.
    :type name: str
    :param token: Authentication token created for your account.
    :type token: str
    :param url: Vision One API url this client connects to.
    :type url: str
    :param pool_connections: (optional) Number of connection to cache.
    :type pool_connections: int
    :param pool_maxsize: (optional) Maximum size of the pool.
    :type pool_maxsize: int
    :param connect_timeout: (optional) Seconds before connection timeout.
    :type connect_timeout: int
    :param read_timeout: (optional) Seconds before read timeout.
    :type connect_timeout: int
    :rtype: Client
    """
    log.debug(
        "Initializing new client with [Appname=%s, Token=*****, URL=%s]",
        name,
        url,
    )
    return Client(
        Core(
            name,
            token,
            url,
            pool_connections,
            pool_maxsize,
            connect_timeout,
            read_timeout,
        )
    )


@lru_cache(maxsize=None)
class Client:
    def __init__(self, core: Core):
        self._core = core

    def add_alert_note(
        self, alert_id: str, note: str
    ) -> Result[AddAlertNoteResp]:
        """Adds a note to the specified Workbench alert.

        :param alert_id: Workbench alert id.
        :type alert_id: str
        :param note: Value of the note.
        :type note: str
        :rtype: Result[AddAlertNoteResp]:
        """
        return self._core.send(
            AddAlertNoteResp,
            Api.ADD_ALERT_NOTE.value.format(alert_id),
            HttpMethod.POST,
            json={"content": note},
        )

    def add_to_block_list(
        self, *objects: ObjectTask
    ) -> MultiResult[MultiResp]:
        """Adds object(s) to the Suspicious Object List,
        which blocks the objects on subsequent detections.

        :param objects: Object(s) to add.
        :type objects: Tuple[ObjectTask, ...]
        :rtype: MultiResult[MultiResp]
        """
        return self._core.send_multi(
            MultiResp,
            Api.ADD_TO_BLOCK_LIST,
            json=utils.build_object_request(*objects),
        )

    def add_to_exception_list(
        self, *objects: ObjectTask
    ) -> MultiResult[MultiResp]:
        """Adds object(s) to the Exception List.

        :param objects: Object(s) to add.
        :type objects: Tuple[ObjectTask, ...]
        :rtype: MultiResult[MultiResp]
        """
        return self._core.send_multi(
            MultiResp,
            Api.ADD_TO_EXCEPTION_LIST,
            json=utils.build_object_request(*objects),
        )

    def add_to_suspicious_list(
        self, *objects: SuspiciousObjectTask
    ) -> MultiResult[MultiResp]:
        """Adds object(s) to the Suspicious Object List.

        :param objects: Object(s) to add.
        :type objects: Tuple[SuspiciousObjectTask, ...]
        :rtype: MultiResult[MultiResp]
        """
        return self._core.send_multi(
            MultiResp,
            Api.ADD_TO_SUSPICIOUS_LIST,
            json=utils.build_suspicious_request(*objects),
        )

    def collect_file(self, *files: FileTask) -> MultiResult[MultiResp]:
        """Collects a file from one or more endpoints and then sends the files
        to Vision One in a password-protected archive.

        :param files: File(s) to collect.
        :type files: Tuple[FileTask, ...]
        :rtype: MultiResult[MultiResp]
        """
        return self._core.send_endpoint(Api.COLLECT_ENDPOINT_FILE, *files)

    def consume_alert_list(
        self,
        consumer: Callable[[Union[SaeAlert, TiAlert]], None],
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
    ) -> Result[ConsumeLinkableResp]:
        """Retrieves and consume workbench alerts.

        :param start_time: Date that indicates the start of the data retrieval
        time range (yyyy-MM-ddThh:mm:ssZ in UTC).
        Defaults to 24 hours before the request is made.
        :type start_time: Optional[str]
        :param end_time: Date that indicates the end of the data retrieval
        time range (yyyy-MM-ddThh:mm:ssZ in UTC).
        Defaults to the time the request is made.
        :type end_time: Optional[str]
        :param consumer: Function which will consume every record in result.
        :type consumer: Callable[[Union[SaeAlert, TiAlert]], None]
        :rtype: Result[ConsumeLinkableResp]:
        """
        return self._core.send_linkable(
            GetAlertListResp,
            Api.GET_ALERT_LIST,
            consumer,
            params=utils.filter_none(
                {"startDateTime": start_time, "endDateTime": end_time}
            ),
        )

    def consume_endpoint_data(
        self,
        consumer: Callable[[Endpoint], None],
        op: QueryOp,
        *values: str,
    ) -> Result[ConsumeLinkableResp]:
        """Retrieves and consume endpoints.

        :param consumer: Function which will consume every record in result.
        :type consumer: Callable[[Endpoint], None]
        :param op: Query operator to apply.
        :type op: QueryOp
        :param values: Agent guid, login account, endpoint name, ip address,
        mac address, operating system, product code.
        :type values: Tuple[str, ...]
        :rtype: Result[ConsumeLinkableResp]:
        """
        return self._core.send_linkable(
            GetEndpointDataResp,
            Api.GET_ENDPOINT_DATA,
            consumer,
            headers=utils.endpoint_query(op, *values),
        )

    def consume_exception_list(
        self, consumer: Callable[[ExceptionObject], None]
    ) -> Result[ConsumeLinkableResp]:
        """Retrieves and consume exception objects.

        :param consumer: Function which will consume every record in result.
        :type consumer: Callable[[ExceptionObject], None]
        :rtype: Result[ConsumeLinkableResp]:
        """
        return self._core.send_linkable(
            GetExceptionListResp, Api.GET_EXCEPTION_LIST, consumer
        )

    def consume_suspicious_list(
        self, consumer: Callable[[SuspiciousObject], None]
    ) -> Result[ConsumeLinkableResp]:
        """Retrieves and consume suspicious objects.

        :param consumer: Function which will consume every record in result.
        :type consumer: Callable[[SuspiciousObject], None]
        :rtype: Result[ConsumeLinkableResp]:
        """
        return self._core.send_linkable(
            GetSuspiciousListResp, Api.GET_SUSPICIOUS_LIST, consumer
        )

    def delete_email_message(
        self, *messages: Union[EmailMessageUIdTask, EmailMessageIdTask]
    ) -> MultiResult[MultiResp]:
        """Deletes a message from one or more mailboxes.

        :param messages: Message(s) to delete.
        :type messages: Tuple[EmailUIdTask, EmailMsgIdTask, ...]
        :rtype: MultiResult[MultiResp]
        """
        return self._core.send_multi(
            MultiResp,
            Api.DELETE_EMAIL_MESSAGE,
            json=[
                task.dict(by_alias=True, exclude_none=True)
                for task in messages
            ],
        )

    def disable_account(
        self, *accounts: AccountTask
    ) -> MultiResult[MultiResp]:
        """Signs the user out of all active application and browser sessions,
        and prevents the user from signing in any new session.

        :param accounts: Account(s) to disable.
        :type accounts: Tuple[AccountTask, ...]
        :rtype: MultiResult[MultiResp]
        """
        return self._core.send_multi(
            MultiResp,
            Api.DISABLE_ACCOUNT,
            json=[
                task.dict(by_alias=True, exclude_none=True)
                for task in accounts
            ],
        )

    def download_sandbox_analysis_result(
        self,
        submit_id: str,
        poll: bool = True,
        poll_time_sec: float = 1800,
    ) -> Result[BytesResp]:
        """Downloads the analysis results of the specified object as PDF.

        :param submit_id: Sandbox submission id.
        :type submit_id: str
        :param poll: If we should wait until the task is finished before
        to return the result.
        :type poll: bool
        :param poll_time_sec: Maximum time to wait for the result to
         be available.
        :type poll_time_sec: float
        :rtype: Result[BytesResp]:
        """
        return self._core.send_sandbox_result(
            BytesResp,
            Api.DOWNLOAD_SANDBOX_ANALYSIS_RESULT,
            submit_id,
            poll,
            poll_time_sec,
        )

    def download_sandbox_investigation_package(
        self,
        submit_id: str,
        poll: bool = True,
        poll_time_sec: float = 1800,
    ) -> Result[BytesResp]:
        """Downloads the Investigation Package of the specified object.

        :param submit_id: Sandbox submission id.
        :type submit_id: str
        :param poll: If we should wait until the task is finished before
        to return the result.
        :type poll: bool
        :param poll_time_sec: Maximum time to wait for the result to
         be available.
        :type poll_time_sec: float
        :rtype: Result[BytesResp]:
        """
        return self._core.send_sandbox_result(
            BytesResp,
            Api.DOWNLOAD_SANDBOX_INVESTIGATION_PACKAGE,
            submit_id,
            poll,
            poll_time_sec,
        )

    def edit_alert_status(
        self,
        alert_id: str,
        status: InvestigationStatus,
        if_match: str,
    ) -> Result[NoContentResp]:
        """Edit the status of an alert or investigation triggered in Workbench.

        :param alert_id: Workbench alert id.
        :type alert_id: str
        :param status: Status to be updated.
        :type status: InvestigationStatus
        :param if_match: Target resource will be updated only if
         it matches ETag of the target one.
        :type if_match: str
        :rtype: Result[NoContentResp]:
        """
        return self._core.send(
            NoContentResp,
            Api.EDIT_ALERT_STATUS.value.format(alert_id),
            HttpMethod.PATCH,
            json={"investigationStatus": status},
            headers={
                "If-Match": (
                    if_match
                    if if_match.startswith('"')
                    else '"' + if_match + '"'
                )
            },
        )

    def enable_account(self, *accounts: AccountTask) -> MultiResult[MultiResp]:
        """Allows the user to sign in to new application and browser sessions.

        :param accounts: Account(s) to enable.
        :type accounts: Tuple[AccountTask, ...]
        :rtype: MultiResult[MultiResp]
        """
        return self._core.send_multi(
            MultiResp,
            Api.ENABLE_ACCOUNT,
            json=[
                task.dict(by_alias=True, exclude_none=True)
                for task in accounts
            ],
        )

    def get_alert_details(self, alert_id: str) -> Result[GetAlertDetailsResp]:
        """Displays information about the specified alert.

        :param alert_id: Workbench alert id.
        :type alert_id: str
        :rtype: Result[GetAlertDetailsResp]:
        """
        return self._core.send(
            GetAlertDetailsResp,
            Api.GET_ALERT_DETAILS.value.format(alert_id),
        )

    def get_alert_list(
        self, start_time: Optional[str] = None, end_time: Optional[str] = None
    ) -> Result[GetAlertListResp]:
        """Retrieves workbench alerts in a paginated list.

        :param start_time: Date that indicates the start of the data retrieval
        time range (yyyy-MM-ddThh:mm:ssZ in UTC).
        Defaults to 24 hours before the request is made.
        :type start_time: Optional[str]
        :param end_time: Date that indicates the end of the data retrieval
        time range (yyyy-MM-ddThh:mm:ssZ in UTC).
        Defaults to the time the request is made.
        :type end_time: Optional[str]
        :rtype: Result[GetAlertListResp]:
        """
        return self._core.send(
            GetAlertListResp,
            Api.GET_ALERT_LIST,
            params=utils.filter_none(
                {"startDateTime": start_time, "endDateTime": end_time}
            ),
        )

    def get_base_task_result(
        self,
        task_id: str,
        poll: bool = True,
        poll_time_sec: float = 1800,
    ) -> Result[BaseTaskResp]:
        """Retrieves the result of a response task.

        :param task_id: Task id.
        :type task_id: str
        :param poll: If we should wait until the task is finished before
         to return the result.
        :type poll: bool
        :param poll_time_sec: Maximum time to wait for the result
        to be available.
        :type poll_time_sec: float
        :rtype: Result[BaseTaskResultResp]:
        """
        return self._core.send_task_result(
            BaseTaskResp, task_id, poll, poll_time_sec
        )

    def get_endpoint_data(
        self, op: QueryOp, *values: str
    ) -> Result[GetEndpointDataResp]:
        """Retrieves endpoints in a paginated list filtered by provided values.

        :param op: Query operator to apply.
        :type op: QueryOp
        :param values: Agent guid, login account, endpoint name, ip address,
        mac address, operating system, product code.
        :type values: Tuple[str, ...]
        :rtype: Result[GetEndpointDataResp]:
        """
        return self._core.send(
            GetEndpointDataResp,
            Api.GET_ENDPOINT_DATA,
            headers=utils.endpoint_query(op, *values),
        )

    def get_exception_list(self) -> Result[GetExceptionListResp]:
        """Retrieves exception objects in a paginated list.

        :rtype: Result[GetExceptionListResp]:
        """
        return self._core.send(GetExceptionListResp, Api.GET_EXCEPTION_LIST)

    def get_sandbox_analysis_result(
        self,
        submit_id: str,
        poll: bool = True,
        poll_time_sec: float = 1800,
    ) -> Result[SandboxAnalysisResultResp]:
        """Retrieves the analysis results of the specified object.

        :param submit_id: Sandbox submission id.
        :type submit_id: str
        :param poll: If we should wait until the task is finished before
         to return the result.
        :type poll: bool
        :param poll_time_sec: Maximum time to wait for the result
         to be available.
        :type poll_time_sec: float
        :rtype: Result[SandboxAnalysisResultResp]:
        """
        return self._core.send_sandbox_result(
            SandboxAnalysisResultResp,
            Api.GET_SANDBOX_ANALYSIS_RESULT,
            submit_id,
            poll,
            poll_time_sec,
        )

    def get_sandbox_submission_status(
        self, submit_id: str
    ) -> Result[SandboxSubmissionStatusResp]:
        """Retrieves the submission status of the specified object.

        :param submit_id: Sandbox submission id.
        :type submit_id: str
        :rtype: Result[SandboxSubmissionStatusResp]:
        """
        return self._core.send(
            SandboxSubmissionStatusResp,
            Api.GET_SANDBOX_SUBMISSION_STATUS.value.format(submit_id),
        )

    def get_sandbox_suspicious_list(
        self,
        submit_id: str,
        poll: bool = True,
        poll_time_sec: float = 1800,
    ) -> Result[SandboxSuspiciousListResp]:
        """Retrieves the suspicious object list associated to the
        specified object.

        :param submit_id: Sandbox submission id.
        :type submit_id: str
        :param poll: If we should wait until the task is finished before
         to return the result.
        :type poll: bool
        :param poll_time_sec: Maximum time to wait for the result
         to be available.
        :type poll_time_sec: float
        :rtype: Result[SandboxSuspiciousListResp]:
        """
        return self._core.send_sandbox_result(
            SandboxSuspiciousListResp,
            Api.GET_SANDBOX_SUSPICIOUS_LIST,
            submit_id,
            poll,
            poll_time_sec,
        )

    def get_suspicious_list(
        self,
    ) -> Result[GetSuspiciousListResp]:
        """Retrieves suspicious objects in a paginated list.

        :rtype: Result[GetSuspiciousListResp]:
        """
        return self._core.send(GetSuspiciousListResp, Api.GET_SUSPICIOUS_LIST)

    def get_task_result(
        self,
        task_id: str,
        class_: Type[S],
        poll: bool = True,
        poll_time_sec: float = 1800,
    ) -> Result[S]:
        """Retrieves the result of a response task.

        :param task_id: Task id.
        :type task_id: str
        :param class_: Expected task result class.
        :type class_: Type[S]
        :param poll: If we should wait until the task is finished before
         to return the result.
        :type poll: bool
        :param poll_time_sec: Maximum time to wait for the result
        to be available.
        :type poll_time_sec: float
        :rtype: Result[BaseTaskResultResp]:
        """
        return self._core.send_task_result(
            class_, task_id, poll, poll_time_sec
        )

    def isolate_endpoint(
        self, *endpoints: EndpointTask
    ) -> MultiResult[MultiResp]:
        """Disconnects one or more endpoints from the network
        but allows communication with the managing Trend Micro server product.

        :param endpoints: Endpoint(s) to isolate.
        :type endpoints: Tuple[EndpointTask, ...]
        :rtype: MultiResult[MultiResp]
        """
        return self._core.send_endpoint(Api.ISOLATE_ENDPOINT, *endpoints)

    def quarantine_email_message(
        self, *messages: Union[EmailMessageUIdTask, EmailMessageIdTask]
    ) -> MultiResult[MultiResp]:
        """Quarantine a message from one or more mailboxes.

        :param messages: Message(s) to quarantine.
        :type messages: Tuple[EmailUIdTask, EmailMsgIdTask, ...]
        :rtype: MultiResult[MultiResp]
        """
        return self._core.send_multi(
            MultiResp,
            Api.QUARANTINE_EMAIL_MESSAGE,
            json=[
                task.dict(by_alias=True, exclude_none=True)
                for task in messages
            ],
        )

    def remove_from_block_list(
        self, *objects: ObjectTask
    ) -> MultiResult[MultiResp]:
        """Removes object(s) that was added to the Suspicious Object List
          using the "Add to block list" action

        :param objects: Object(s) to remove.
        :type objects: Tuple[ObjectTask, ...]
        :rtype: MultiResult[MultiResp]
        """
        return self._core.send_multi(
            MultiResp,
            Api.REMOVE_FROM_BLOCK_LIST,
            json=utils.build_object_request(*objects),
        )

    def remove_from_exception_list(
        self, *objects: ObjectTask
    ) -> MultiResult[MultiResp]:
        """Removes object(s) from the Exception List.

        :param objects: Object(s) to remove.
        :type objects: Tuple[ObjectTask, ...]
        :rtype: MultiResult[MultiResp]
        """
        return self._core.send_multi(
            MultiResp,
            Api.REMOVE_FROM_EXCEPTION_LIST,
            json=utils.build_object_request(*objects),
        )

    def remove_from_suspicious_list(
        self, *objects: ObjectTask
    ) -> MultiResult[MultiResp]:
        """Removes object(s) from the Suspicious List.

        :param objects: Object(s) to remove.
        :type objects: Tuple[ObjectTask, ...]
        :rtype: MultiResult[MultiResp]
        """
        return self._core.send_multi(
            MultiResp,
            Api.REMOVE_FROM_SUSPICIOUS_LIST,
            json=utils.build_object_request(*objects),
        )

    def reset_password_account(
        self, *accounts: AccountTask
    ) -> MultiResult[MultiResp]:
        """Signs the user out of all active application and browser sessions,
        and forces the user to create a new password during the next sign-in
        attempt.

        :param accounts: Account(s) to reset.
        :type accounts: Tuple[AccountTask, ...]
        :rtype: MultiResult[MultiResp]
        """
        return self._core.send_multi(
            MultiResp,
            Api.RESET_PASSWORD,
            json=[
                task.dict(by_alias=True, exclude_none=True)
                for task in accounts
            ],
        )

    def restore_endpoint(
        self, *endpoints: EndpointTask
    ) -> MultiResult[MultiResp]:
        """Restores network connectivity to one or more endpoints that applied
        the "Isolate endpoint" action.

        :param endpoints: Endpoint(s) to restore.
        :type endpoints: Tuple[EndpointTask, ...]
        :rtype: MultiResult[MultiResp]
        """
        return self._core.send_endpoint(Api.RESTORE_ENDPOINT, *endpoints)

    def restore_email_message(
        self, *messages: Union[EmailMessageUIdTask, EmailMessageIdTask]
    ) -> MultiResult[MultiResp]:
        """Restore quarantined email message(s).

        :param messages: Message(s) to restore.
        :type messages: Tuple[EmailUIdTask, EmailMsgIdTask, ...]
        :rtype: MultiResult[MultiResp]
        """
        return self._core.send_multi(
            MultiResp,
            Api.RESTORE_EMAIL_MESSAGE,
            json=[
                task.dict(by_alias=True, exclude_none=True)
                for task in messages
            ],
        )

    def sign_out_account(
        self, *accounts: AccountTask
    ) -> MultiResult[MultiResp]:
        """Signs the user out of all active application and browser sessions.

        :param accounts: Account(s) to sign out.
        :type accounts: Tuple[AccountTask, ...]
        :rtype: MultiResult[MultiResp]
        """
        return self._core.send_multi(
            MultiResp,
            Api.SIGN_OUT_ACCOUNT,
            json=[
                task.dict(by_alias=True, exclude_none=True)
                for task in accounts
            ],
        )

    def submit_file_to_sandbox(
        self,
        file: bytes,
        file_name: str,
        document_password: Optional[str] = None,
        archive_password: Optional[str] = None,
        arguments: Optional[str] = None,
    ) -> Result[SubmitFileToSandboxResp]:
        """Submits a file to the sandbox for analysis.

        :param file: Raw content in bytes.
        :type file: bytes
        :param file_name: Name of the file.
        :type file_name: str
        :param document_password: Password used to
         decrypt the submitted file sample.
        :type document_password: Optional[str]
        :param archive_password: Password encoded in Base64 used to decrypt
         the submitted archive.
        :type archive_password: Optional[str]
        :param arguments: Command line arguments to run the submitted file.
         Only available for Portable Executable (PE) files and script files.
        :type arguments: Optional[str]
        :rtype: Result[SubmitFileToSandboxResp]:
        """
        return self._core.send(
            SubmitFileToSandboxResp,
            Api.SUBMIT_FILE_TO_SANDBOX,
            HttpMethod.POST,
            data=utils.build_sandbox_file_request(
                document_password, archive_password, arguments
            ),
            files={"file": (file_name, file, "application/octet-stream")},
        )

    def submit_urls_to_sandbox(self, *urls: str) -> MultiResult[MultiUrlResp]:
        """Submits URLs to the sandbox for analysis.

        :param urls: URL(s) to be submitted.
        :type urls: Tuple[str, ...]
        :rtype: MultiResult[MultiUrlResp]
        """
        return self._core.send_multi(
            MultiUrlResp,
            Api.SUBMIT_URLS_TO_SANDBOX,
            json=[{"url": url} for url in urls],
        )

    def terminate_process(
        self, *processes: ProcessTask
    ) -> MultiResult[MultiResp]:
        """Terminates a process that is running on one or more endpoints.

        :param processes: Process(es) to terminate.
        :type processes: Tuple[ProcessTask, ...]
        :rtype: MultiResult[MultiResp]
        """
        return self._core.send_endpoint(
            Api.TERMINATE_ENDPOINT_PROCESS, *processes
        )

    def check_connectivity(self) -> Result[ConnectivityResp]:
        """Checks the connection to the API service
        and verifies if your authentication token is valid.

        :rtype: Result[ConnectivityResp]
        """
        return self._core.send(ConnectivityResp, Api.CONNECTIVITY)



from typing import Dict, List

from pydantic.utils import to_lower_camel

from .model.commons import (
    Alert,
    Entity,
    HostInfo,
    Indicator,
    SaeAlert,
    TiAlert,
)

INDICATOR_CEF_MAP: Dict[str, str] = {
    "command_line": "dproc",
    "url": "request",
    "domain": "sntdom",
    "ip": "src",
    "email_sender": "suser",
    "fullpath": "filePath",
    "filename": "fname",
    "file_sha1": "fileHash",
    "user_account": "suser",
    "host": "shost",
    "port": "spt",
    "process_id": "dpid",
    "registry_key": "TrendMicroVoRegistryKeyHandle",
    "registry_value": "TrendMicroVoRegistryValue",
    "registry_value_data": "TrendMicroVoRegistryData",
    "file_sha256": "TrendMicroVoFileHashSha256",
    "email_message_id": "TrendMicroVoEmailMessageId",
    "email_message_unique_id": "TrendMicroVoEmailMessageUniqueId",
}


def map_cef(alert: Alert) -> Dict[str, str]:
    data: Dict[str, str] = _map_common(alert)
    _map_entities(data, alert.impact_scope.entities)
    _map_indicators(data, alert.indicators)
    if isinstance(alert, SaeAlert):
        _map_sae(data, alert)
    if isinstance(alert, TiAlert):
        _map_ti(data, alert)
    return data


def _map_common(alert: Alert) -> Dict[str, str]:
    return dict(
        externalId=alert.id,
        act=alert.investigation_status,
        cat=alert.model,
        Severity=alert.severity,
        rt=alert.created_date_time,
        sourceServiceName=alert.alert_provider,
        msg="Workbench Link: " + alert.workbench_link,
        cnt=str(alert.score),
        cn1=str(alert.impact_scope.desktop_count),
        cn1Label="Desktop Count",
        cn2=str(alert.impact_scope.server_count),
        cn2Label="Server Count",
        cn3=str(alert.impact_scope.account_count),
        cn3Label="Account Count",
        cn4=str(alert.impact_scope.email_address_count),
        cn4Label="Email Address Count",
        cs1=", ".join(alert.indicators[0].provenance),
        cs1Label="Provenance",
    )


def _map_entities(data: Dict[str, str], entities: List[Entity]) -> None:
    for entity in entities:
        if isinstance(entity.entity_value, HostInfo):
            data["dhost"] = entity.entity_value.name
            data["dst"] = ", ".join(entity.entity_value.ips)
        else:
            data["duser"] = entity.entity_value


def _map_indicators(data: Dict[str, str], indicators: List[Indicator]) -> None:
    for indicator in indicators:
        if isinstance(indicator.value, HostInfo):
            data["shost"] = indicator.value.name
            data["src"] = ", ".join(indicator.value.ips)
        else:
            data[
                INDICATOR_CEF_MAP.get(
                    indicator.type, to_lower_camel(indicator.type)
                )
            ] = indicator.value


def _map_sae(data: Dict[str, str], alert: SaeAlert) -> None:
    data["cs2"] = alert.matched_rules[0].matched_filters[0].name
    data["cs2Label"] = "Matched Filter"
    data["cs3"] = ", ".join(
        alert.matched_rules[0].matched_filters[0].mitre_technique_ids
    )
    data["cs3Label"] = "Matched Techniques"
    data["reason"] = alert.matched_rules[0].name
    data["msg"] = data.get("msg", "") + f"\nDescription: {alert.description}"


def _map_ti(data: Dict[str, str], alert: TiAlert) -> None:
    data["cs2"] = ", ".join(alert.matched_indicator_patterns[0].tags)
    data["cs2Label"] = "Matched Pattern Tags"
    data["cs3"] = alert.matched_indicator_patterns[0].pattern
    data["cs3Label"] = "Matched Pattern"
    data["msg"] = data.get("msg", "") + f"\nReport Link: {alert.report_link}"
    data["createdBy"] = alert.created_by
    if alert.campaign:
        data["campaign"] = alert.campaign
    if alert.industry:
        data["industry"] = alert.industry
    if alert.region_and_country:
        data["regionAndCountry"] = alert.region_and_country



from .__about__ import __version__
from .caller import Client, client
from .mapper import map_cef
from .model.commons import (
    Alert,
    Digest,
    Endpoint,
    Entity,
    Error,
    ExceptionObject,
    HostInfo,
    ImpactScope,
    Indicator,
    MatchedEvent,
    MatchedFilter,
    MatchedIndicatorPattern,
    MatchedRule,
    MsData,
    MsDataUrl,
    MsError,
    SaeAlert,
    SaeIndicator,
    SandboxSuspiciousObject,
    SuspiciousObject,
    TiAlert,
    TiIndicator,
    Value,
    ValueList,
)
from .model.enums import (
    EntityType,
    InvestigationStatus,
    ObjectType,
    OperatingSystem,
    ProductCode,
    Provenance,
    Provider,
    QueryField,
    QueryOp,
    RiskLevel,
    SandboxAction,
    SandboxObjectType,
    ScanAction,
    Severity,
    Status,
    TaskAction,
)
from .model.requests import (
    AccountTask,
    EmailMessageIdTask,
    EmailMessageUIdTask,
    EndpointTask,
    FileTask,
    ObjectTask,
    ProcessTask,
    SuspiciousObjectTask,
)
from .model.responses import (
    AccountTaskResp,
    AddAlertNoteResp,
    BaseTaskResp,
    BlockListTaskResp,
    BytesResp,
    CollectFileTaskResp,
    ConnectivityResp,
    ConsumeLinkableResp,
    EmailMessageTaskResp,
    EndpointTaskResp,
    GetAlertDetailsResp,
    GetAlertListResp,
    GetEndpointDataResp,
    GetExceptionListResp,
    GetSuspiciousListResp,
    MultiResp,
    MultiUrlResp,
    NoContentResp,
    SandboxAnalysisResultResp,
    SandboxSubmissionStatusResp,
    SandboxSubmitUrlTaskResp,
    SandboxSuspiciousListResp,
    SubmitFileToSandboxResp,
    TerminateProcessTaskResp,
)
from .results import MultiResult, Result, ResultCode

__all__ = [
    "__version__",
    "client",
    "map_cef",
    "AccountTask",
    "AccountTaskResp",
    "AddAlertNoteResp",
    "Alert",
    "BaseTaskResp",
    "BlockListTaskResp",
    "BytesResp",
    "Client",
    "CollectFileTaskResp",
    "ConnectivityResp",
    "ConsumeLinkableResp",
    "Digest",
    "EmailMessageIdTask",
    "EmailMessageTaskResp",
    "EmailMessageUIdTask",
    "Endpoint",
    "EndpointTask",
    "EndpointTaskResp",
    "Entity",
    "EntityType",
    "Error",
    "ExceptionObject",
    "FileTask",
    "GetAlertDetailsResp",
    "GetAlertListResp",
    "GetEndpointDataResp",
    "GetExceptionListResp",
    "GetSuspiciousListResp",
    "HostInfo",
    "ImpactScope",
    "Indicator",
    "InvestigationStatus",
    "MatchedEvent",
    "MatchedFilter",
    "MatchedIndicatorPattern",
    "MatchedRule",
    "MsData",
    "MsDataUrl",
    "MsError",
    "MultiResult",
    "MultiResp",
    "MultiUrlResp",
    "NoContentResp",
    "ObjectTask",
    "ObjectType",
    "OperatingSystem",
    "ProcessTask",
    "ProductCode",
    "Provenance",
    "Provider",
    "QueryField",
    "QueryOp",
    "Result",
    "ResultCode",
    "RiskLevel",
    "SaeAlert",
    "SaeIndicator",
    "SandboxAction",
    "SandboxAnalysisResultResp",
    "SandboxObjectType",
    "SandboxSubmissionStatusResp",
    "SandboxSubmitUrlTaskResp",
    "SandboxSuspiciousListResp",
    "SandboxSuspiciousObject",
    "ScanAction",
    "Severity",
    "Status",
    "SubmitFileToSandboxResp",
    "SuspiciousObject",
    "SuspiciousObjectTask",
    "TaskAction",
    "TerminateProcessTaskResp",
    "TiAlert",
    "TiIndicator",
    "Value",
    "ValueList",
]



from typing import Optional

from .commons import BaseModel
from .enums import ObjectType, RiskLevel, ScanAction


class AccountTask(BaseModel):
    account_name: str
    """User account name."""
    description: Optional[str] = None
    """Description of a response task."""


class EndpointTask(BaseModel):
    endpoint_name: Optional[str]
    """Endpoint name."""
    agent_guid: Optional[str] = None
    """Agent guid"""
    description: Optional[str] = None
    """Description of a response task."""


class EmailMessageIdTask(BaseModel):
    message_id: str
    """Email message id."""
    mail_box: Optional[str]
    """Email address."""
    description: Optional[str] = None
    """Description of a response task."""


class EmailMessageUIdTask(BaseModel):
    unique_id: str
    """Email unique message id."""
    description: Optional[str] = None
    """Description of a response task."""


class ObjectTask(BaseModel):
    object_type: ObjectType
    """Type of object."""
    object_value: str
    """Value of an object."""
    description: Optional[str] = None
    """Description of an object."""


class SuspiciousObjectTask(ObjectTask):
    scan_action: Optional[ScanAction] = None
    """Action applied after detecting a suspicious object."""
    risk_level: Optional[RiskLevel] = None
    """Risk level of a suspicious object."""
    days_to_expiration: Optional[int] = None
    """Number of days before the object expires."""


class FileTask(EndpointTask):
    file_path: str
    """File path of the file to be collected from the target."""


class ProcessTask(EndpointTask):
    file_sha1: str
    """SHA1 hash of the terminated process's executable file."""
    file_name: Optional[str] = None
    """File name of the target."""



from __future__ import annotations

from typing import List

from requests import Response

from .model.commons import Error, MsError


class ServerCustError(Exception):
    def __init__(self, status: int, message: str):
        super().__init__(message)
        self.status = status


class ServerJsonError(Exception):
    def __init__(self, error: Error):
        super().__init__(
            f"Error response received from Vision One. [Error={error}]"
        )
        self.error = error


class ServerMultiJsonError(Exception):
    def __init__(self, errors: List[MsError]):
        super().__init__(
            (
                "Multi error response received from Vision One."
                f" [Errors={errors}]"
            ),
        )
        self.errors = errors


class ServerHtmlError(ServerCustError):
    def __init__(self, status: int, html: str):
        super().__init__(
            status,
            html,
        )


class ServerTextError(ServerCustError):
    def __init__(self, status: int, text: str):
        super().__init__(
            status,
            text,
        )


class ParseModelError(ServerCustError):
    def __init__(self, model: str, raw_response: Response):
        super().__init__(
            500,
            (
                "Could not parse response from Vision One.\n"
                f"Conditions unmet [Model={model}, {raw_response}]"
            ),
        )
